------------------------
-- Kevin Depue (2010) --
-- Backflip Studios   --
--                    --
-- LuaMainGame        --
------------------------

-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local pool      = chunks.pool
local utility   = chunks.utility
local scenarios = chunks.scenarios
local assets    = chunks.assets
local events    = chunks.events

-- LuaMainGame --
LuaMainGame = chunks.class.new
{
	-- actors --
	root = nil,
	
	-- walls --
	speed = -340.0,
	
	-- height --
	height_factor  = 10.0,
	height         = 0.0,
	best_height    = 0.0,
	spawned_banner = false,
	
	-- shield --
	shield_delay_range = v2f( 15.0, 20.0 ),
	shield_timer       = 0.0,
	
	-- scenarios --
	scenario_root  = nil,
	scenario_timer = 0.0,
	foliage_timer  = 0.0,
	
	-- bonus --
	bonus_timer      = 0.0,
	bonus_extra_time = 0.5,
	bonus_empty_time = 2.25,
	
	-- references --
	background = nil,
	walls      = nil,
	ninja      = nil,
	hud        = nil
}

-- The spawner.
spawn = chunks.class.spawner( MainGame, LuaMainGame )

-- The created method.
function LuaMainGame:created( game )
	-- Set clearing info.
	game.clear_color    = constants.clear_color
	game.clear_color_on = constants.clear_color_on
	game.clear_depth_on = constants.clear_depth_on
	
	-- Spawn root.
	self.root = game:getScene():spawn( ActorBase() )
	
	-- Scene camera.
	do
		local camera = game:getScene():spawn( OrthographicCamera() )
		
		camera.viewport_offset = constants.scene.viewport_offset
		camera.viewport_size   = constants.scene.viewport_cropped
		--camera.viewport_size   = constants.scene.viewport_size
		camera.screen_size     = constants.scene.screen_size
		camera.world_size      = constants.scene.world_size
		camera.clipping_planes = constants.scene.clipping_planes
		
		game.scene_camera = camera
		
		--camera.zoom_factor = 0.6
	end
	
	-- Window camera.
	do
		local camera = game:getWindow():spawn( OrthographicCamera() )
		
		camera.viewport_offset = constants.window.viewport_offset
		camera.viewport_size   = constants.window.viewport_size
		camera.screen_size     = constants.window.screen_size
		camera.world_size      = constants.window.adjusted_size --world_size
		camera.clipping_planes = constants.window.clipping_planes
		
		game.window_camera = camera
	end
	
	-- Set best height.
	self.best_height = UserData.get():getHighScore( engine.theme )
	
	-- Setup scenario root.
	self.scenario_root = self.root:spawn( ActorBase() )
	
	-- Spawn the background.
	self.background = self.root:spawn( chunks.background.spawn() )
	
	-- Spawn the walls.
	self.walls = self.root:spawn( chunks.walls.spawn() )
	
	-- Spawn the ninja.
	self.ninja = self.root:spawn( chunks.ninja.spawn() )
	
	-- Spawn the foreground.
	self.root:spawn( chunks.foreground.spawn() )
	
	-- Spawn the hud.
	self.hud = game:getWindow():spawn( chunks.hud.spawn() )
	
	--game.delta_time_factor = 0.5
	
	-- Start music.
	do
		local music = Audio.get().music
		
		music:clearPlaylist()
		music:addToPlaylist( assets.music.intro )
		music:addToPlaylist( assets.music.loop )
		music:play()
	end
	
	local gameSelf = self;

	engine.onMenuButton = function()
		engine.onMenuButton = function() end;
		gameSelf:onPause()
	end

	engine.onBackButton = function()
		engine.onBackButton = function() end;
		gameSelf:onPause()
--		engine:spawnMenuGame()
--		game:destroy()
	end

	self.onMenuButton = engine.onMenuButton;
	self.onBackButton = engine.onBackButton;

	-- Analytics.
	events.on_start_game( game )
end

-- The tick event.
function LuaMainGame:tick( game, delta_time )
	local adjusted = self.ninja.lua.bonus and delta_time * self.ninja.lua.bonus_factor or delta_time
	
	-- If we're not paused and the ninja isn't dying, update things.
	if game.paused == false and self.ninja.lua.dead == false then
		-- Update bonus (use unadjusted delta_time).
		self:updateBonus( delta_time )
		
		-- Update the height.
		self:updateHeight( adjusted )

		-- Update scenarios.
		self:updateScenarios( adjusted )
	end
end

-- Update height.
function LuaMainGame:updateHeight( delta_time )
	self.height = self.height + ( delta_time * math.abs( self.speed ) 
			                  / self.height_factor )
	self.hud.lua.ticker.value = self.height
end

-- Update bonus.
function LuaMainGame:updateBonus( delta_time )
	-- Update bonus timer.
	if self.bonus_timer > 0.0 then
		self.bonus_timer = self.bonus_timer - delta_time
	end
end

-- Choses a scenario, based on the game's height.
function LuaMainGame:chooseScenario()
	local height       = self.height
	local transitions  = scenarios.transitions
	local size         = #transitions
	local allow_shield = self.shield_timer <= 0.0 and not self.ninja.lua.bonus 

	-- First, find the correct transition.
	for i = 1,size do
		local transition = transitions[ i ]

		-- Are we in the right transition?
		if height <= transition.height or i == size then
			-- Keep looping until we find one.
			while true do
				local prob = math.random()
				local s    = transition.scenarios[ math.random( #transition.scenarios ) ]
				
				-- Conditions:
				-- 1) The probability is within range.
				-- 2) Shields must be allowed if a shield is chosen.
				if prob <= s.prob and not ( s.f == scenarios.scenarios.shieldpack and not allow_shield ) then
					return s.f
				end
			end
		end
	end	
end

-- Spawn a scenario.  Returns the amount of
-- time to wait before spawning another scenario.
function LuaMainGame:spawnScenario()
	local game     = self.owner
	local root     = self.scenario_root:spawn( chunks.scenario.spawn() )
	local f        = nil
	local result   = nil
	
	-- Spawn new high score banner?
	if not self.spawned_banner and self.height > self.best_height and self.best_height > 0 then
		f                   = scenarios.scenarios.new_high_score
		self.spawned_banner = true
	-- Force empty scenarios during end of bonus?
	elseif self.bonus_timer > 0.0 and self.bonus_timer <= self.bonus_empty_time then
		f = scenarios.scenarios.empty
	-- Otherwise, randomly choose one.
	else
		f = self:chooseScenario()
	end
	
	-- If a shieldpack was chosen, update timer.
	if f == scenarios.scenarios.shieldpack then
		self.shield_timer = utility.frandom( self.shield_delay_range.x, 
			                                 self.shield_delay_range.y )
	end
	
	-- Spawn the scenario.
	local result = f( root )
	
	-- Randomly mirror the scenario if it's not a new high score banner.
	if f ~= scenarios.scenarios.new_high_score and math.random() < 0.5 then
		root.lua:mirror()
	end
	
	-- Setup cleanup actions.
	do
		local s = SequenceAction()
		
		s:addAction( DestroyAction( result.lifetime ) )
		
		root:addAction( s )
	end
	
	return result.delay
end

function LuaMainGame:spawnFoliage()
	local foliage = scenarios.scenarios.foliage
	
	-- No foliage?
	if foliage == nil then
		return 0.0
	end
	
	local game   = self.owner
	local root   = self.scenario_root:spawn( chunks.scenario.spawn() )
	local result = foliage( root )

	-- Randomly mirror the foliage.
	if math.random() < 0.5 then
		root.lua:mirror()
	end
	
	-- Setup cleanup actions.
	do
		local s = SequenceAction()
		
		s:addAction( DestroyAction( result.lifetime ) )
		
		root:addAction( s )
	end
	
	return result.delay
end

-- Scenarios.
function LuaMainGame:updateScenarios( delta_time )
	-- Update shield timing information.
	if self.shield_timer > 0.0 then
		self.shield_timer = self.shield_timer - delta_time
	end
	
	-- See if we need to spawn another foliage.
	do
		self.foliage_timer = self.foliage_timer - delta_time
		
		if self.foliage_timer <= 0.0 then
			self.foliage_timer = self:spawnFoliage()
		end
	end
		
	-- See if we need to spawn another scenario.
	do
		self.scenario_timer = self.scenario_timer - delta_time
		
		if self.scenario_timer <= 0.0 then
			self.scenario_timer = self:spawnScenario()
		end
	end
end

-- Input events.
function LuaMainGame:onInputEvents( game, events )
	-- If we're paused, exit.
	if self.owner.paused or self.ninja.lua.dead or self.ninja.lua.bonus then
		return
	end
	
	-- Loop through the events.
	for event in events.all do
		local touch = TouchEvent.fast_cast( event )
                
        -- If we tapped, jump.
        if touch and touch.type == TouchEvent.TouchStart then
			self.ninja.lua:jump()
			
			touch.eaten = true
			
			break
		end        
	end
end

-- Intro sequence.
function LuaMainGame:intro( replay )
	local game = self.owner
	
	-- Start paused.
	game.paused = true
	
	-- Spawn a fade in menu.
	do
		local callback = function( inner )
			game.paused = false
		end
		
		if replay then
			chunks.fade.fade_in( self.hud, callback )
		else
			chunks.fade.fade_in_loading( self.hud, callback )
		end
	end
end

-- The game over event.
function LuaMainGame:onGameOver()
	local game = self.owner
	
	game.paused = true
	
	-- Run game over visitor.
	do
		local callback = function( inner )
			if inner.lua and inner.lua.onGameOver then
				inner.lua:onGameOver()
			end
		end
		
		chunks.visitors.visit( game:getScene(), callback )
	end
	
	-- Spawn the death menu.
	self.hud:spawn( chunks.death_menu.spawn() )
	
	-- Update score.
	UserData.get():setHighScore( engine.theme, self.height )
	UserData.get():flush()
end

-- The pause event.
function LuaMainGame:onPause()
	local game = self.owner
	
	if game.paused == false and self.ninja.lua.dead == false then
		game.paused = true
		self.hud.lua.meter.lua:pause()
		self.hud:spawn( chunks.pause_menu.spawn() )
		
		-- Run pause visitor.
		do
			local callback = function( inner )
				if inner.lua and inner.lua.onPause then
					inner.lua:onPause()
				end
			end
			
			chunks.visitors.visit( game:getScene(), callback )
		end
	end
end

-- The unpause event.
function LuaMainGame:onUnpause()
	local game = self.owner

	engine.onMenuButton = self.onMenuButton
	engine.onBackButton = self.onBackButton
	
	if game.paused then
		game.paused = false
		self.hud.lua.meter.lua:unPause()
		
		-- Run unpause visitor.
		do
			local callback = function( inner )
				if inner.lua and inner.lua.onUnpause then
					inner.lua:onUnpause()
				end
			end
			
			chunks.visitors.visit( game:getScene(), callback )
		end
	end
end

-- The quit event.
function LuaMainGame:onQuit()
	local game = self.owner 
	
	-- Spawn a fade out menu.
	do
		local callback = function( inner )
			-- Analytics.
			events.on_quit_game( game )
			
			game:destroy()
			engine:spawnMenuGame()
		end
		
		chunks.fade.fade_out( self.hud, callback )
	end
end

-- The replay event.
function LuaMainGame:onReplay()
	local game = self.owner 
	
	-- Spawn a fade out menu.
	do
		local callback = function( inner )
			game:destroy()
			engine:spawnMainGame( engine.theme, true )
		end
		
		chunks.fade.fade_out( self.hud, callback )
	end
end





